from lurrn.learn import AdaGrad, FileModel, SgdMomentum
from lurrn.feature import FeatureHasher
from lurrn.alphabet import CPPAlphabet, CPPUniAlphabet
from lurrn.sparsmat import SparseVectorsD, SparseVectorD

__all__ = ['FeatureHasher', 'CPPAlphabet', 'CPPUniAlphabet',
           'SparseVectorD', 'SparseVectorsD',
           'all_learners', 'create_learner', 'load_weights']

def all_learners():
    '''
    returns a list of all learner names that can be
    used for calls to create_learner
    '''
    return ['adagrad', 'sgd_momentum']

def create_learner(name, *args, **kwargs):
    '''
    creates a learner object
    '''
    if name == 'adagrad':
        return AdaGrad(*args, **kwargs)
    elif name == 'sgd_momentum':
        return SgdMomentum(*args, **kwargs)
    else:
        return ValueError('no such learner')

def load_weights(fname):
    return FileModel(fname)



import numpy
cimport numpy
from alphabet import PythonAlphabet
from .sparsmat cimport VecD1, SparseVectorD, coordinate_t
cdef extern from "math.h":
    double pow(double x, double y)
    double fabs(double x)
    double sqrt(double x)
    double copysign(double x, double y)

'''
Online learners
---------------

most online learners have two important functions:
- score(vec, is_testing)
  computes a score for the respective vector. If is_testing
  is false, the learner may do some bookkeeping. If is_testing
  is true, the learner should not do that kind of bookkeeping.
- update(vecs)
  updates the vector based on vecs which contains the gradient
  of one partial function
- get_weights()
  retrieves the current weight vector
- get_weights_l1()
  retrieves the l1 norm of the weight vector (may be approximate)
- set_weights(array)
  sets the current weight vector
'''

DEF soften = 1e4

cdef class AdaGrad:
    '''
    inspired by StanfordNLP's SparseAdaGradMinimizer
    but with lazy updates to weight vector
    '''
    # TODO check if this works
    cdef readonly int n_dimensions
    cdef public object fc
    cdef double eta, lambdaL1, lambdaL2
    cdef readonly double[:] weights
    cdef readonly double[:] avgWeights
    cdef double[:] sumGradSquare, lastUpdated
    cdef long int timestep
    cdef bint lazy_updates

    def __init__(self, n_dimensions,
                 l1=0.0, l2=0.0,
                 n_examples=1.0, lazy_updates=False):
        self.weights = numpy.zeros(n_dimensions, 'd')
        self.avgWeights = numpy.zeros(n_dimensions, 'd')
        self.sumGradSquare = numpy.zeros(n_dimensions, 'd')
        self.lastUpdated = numpy.zeros(n_dimensions, 'd')
        self.eta = 0.1
        self.lambdaL1 = l1 / n_examples
        self.lambdaL2 = l2 / n_examples
        self.lazy_updates = lazy_updates
    def set_weights(self, weights):
        if len(weights) != len(self.weights):
            raise ValueError()
        self.weights = weights
        self.avgWeights = self.weights.copy()
    cpdef double score(self, SparseVectorD vec, bint use_avg=True):
        if use_avg:
            return vec._dotFull(& self.avgWeights[0])/self.timestep
        else:
            if self.lazy_updates:
                self.lazyUpdate(vec)
            return vec._dotFull( & self.weights[0])

    def update(self, gradient):
        cdef double a
        cdef int i
        cdef SparseVectorD vec
        cdef double gradf, prev_rate, current_rate, mix_rate, sgsValue
        cdef double wanted_update, trunc1, trunc2, actual_update
        cdef double last_timestep, idle_interval
        cdef coordinate_t feature
        for a, vec in gradient:
            for i from 0 <= i < vec.my_len:
                feature = vec.idx_ptr[i]
                gradf = a * vec.vals_ptr[i]
                prev_rate = self.eta / sqrt(
                    self.sumGradSquare[feature] + soften)
                sgsValue = self.sumGradSquare[feature] + gradf * gradf
                self.sumGradSquare[feature] = sgsValue
                current_rate = self.eta / (sqrt(sgsValue) + soften)
                wanted_update = self.weights[feature] + (current_rate * gradf)
                last_timestep = self.lastUpdated[feature]
                idle_interval = self.timestep - last_timestep - 1
                self.lastUpdated[feature] = self.timestep
                mix_rate = current_rate + prev_rate * idle_interval
                trunc1 = max(0.0, fabs(wanted_update)
                             - mix_rate * self.lambdaL1)
                trunc2 = trunc1 * pow(1 - self.lambdaL2, mix_rate)
                actual_update = copysign(trunc2, wanted_update)
                self.avgWeights[feature] += self.weights[feature] * idle_interval
                self.weights[feature] = actual_update
        self.timestep += 1
    cdef lazyUpdate(self, SparseVectorD vec):
        '''
        performs regularizer updates to weight vector that are to be
        performed lazily
        '''
        cdef int i
        cdef double current_rate, mix_rate, sgsValue
        cdef double last_timestep, idle_interval
        cdef double wanted_update, actual_update
        cdef double trunc1, trunc2
        cdef coordinate_t feature
        for i from 0 <= i < vec.my_len:
            feature = vec.idx_ptr[i]
            last_timestep = self.lastUpdated[feature]
            idle_interval = self.timestep - last_timestep - 1
            if idle_interval > 0:
                wanted_update = self.weights[feature]
                sgsValue = self.sumGradSquare[feature]
                prev_rate = self.eta / sqrt(sgsValue + soften)
                mix_rate = prev_rate * idle_interval
                trunc1 = max(0.0, fabs(wanted_update)
                             - mix_rate * self.lambdaL1)
                trunc2 = trunc1 * pow(1 - self.lambdaL2, mix_rate)
                actual_update = copysign(trunc2, wanted_update)
                self.lastUpdated[feature] = self.timestep
                self.avgWeights[feature] += self.weights[feature] * idle_interval
                self.weights[feature] = actual_update

    def __reduce__(self):
        return (make_classifier, ('adagrad',
                                  numpy.asarray(self.avgWeights)/self.timestep,
                                  self.fc))
    def save_binary(self, fname_out):
        numbers = numpy.asarray(self.avgWeights)/self.timestep
        with file(fname_out, 'w') as f:
            numpy.save(f, numbers)

cdef class SgdMomentum:
    '''
    implements stochastic gradient descent with
    momentum, and L2 regularization based on vector
    scaling (like Bottou's sgdsvm).
    Momentum makes it so that weight updates are spread over
    multiple time points.
    '''
    cdef readonly int n_dimensions
    cdef double eta
    cdef double l2
    cdef int t0
    cdef readonly double[:] weights
    cdef readonly double[:] velocity
    cdef double momentum
    cdef double weights_scale
    cdef readonly long[:] lastUpdated
    cdef long int timestep
    def __init__(self, n_dimensions, l2=0.0,
                 n_examples=1.0, momentum=0.9, eta=0.1):
        self.momentum = momentum
        self.eta = eta
        self.weights = numpy.zeros(n_dimensions, 'd')
        self.velocity = numpy.zeros(n_dimensions, 'd')
        self.lastUpdated = numpy.zeros(n_dimensions, 'l')
        self.weights_scale = 1.0
        self.n_dimensions = n_dimensions
        self.t0 = n_examples
        self.timestep = 0
        self.l2 = l2/n_examples
    cpdef score(self, SparseVectorD vec, is_testing=True):
        self.hiddenUpdates(vec)
        return self.weights_scale * vec._dotFull(& self.weights[0])
    cdef hiddenUpdates(self, SparseVectorD vec, int start=0, int end=-1):
        cdef int i, idle, k
        cdef double mom_factor
        if vec is None:
            if end == -1:
                end = self.n_dimensions
            # perform hidden updates for all parameters
            for k from start <= k < end:
                idle = self.timestep - self.lastUpdated[k]
                mom_l1 = self.momentum / (1.0 - self.l2)
                mom_factor = pow(mom_l1, idle)
                self.weights[k] += 1.0/self.weights_scale * (
                    self.velocity[k] *
                    (1.0 - mom_factor) / (1.0 - mom_l1))
                self.velocity[k] *= pow(self.momentum, idle)
                self.lastUpdated[k] = self.timestep
        else:
            for i from 0 <= i < vec.my_len:
                k = vec.idx_ptr[i]
                idle = self.timestep - self.lastUpdated[k]
                mom_l1 = self.momentum / (1.0 - self.l2)
                mom_factor = pow(mom_l1, idle)
                self.weights[k] += 1.0/self.weights_scale * (
                    self.velocity[k] *
                    (1 - mom_factor) / (1 - mom_l1))
                self.velocity[k] *= pow(self.momentum, idle)
                self.lastUpdated[k] = self.timestep
    def update(self, gradient):
        cdef double a, gradf
        cdef SparseVectorD vec
        cdef coordinate_t i, k
        cdef double eta_actual
        eta_actual = self.eta * self.t0 / (self.t0 + self.timestep)
        for a, vec in gradient:
            for i from 0 <= i < vec.my_len:
                k = vec.idx_ptr[i]
                gradf = a * vec.vals_ptr[i] * eta_actual
                self.velocity[k] += gradf
        self.weights_scale *= (1.0 - eta_actual * self.l2)
        if self.weights_scale < 1e-4:
            print "SGD: rescaling, timestep=%d"%(self.timestep,)
            for k from 0<=k<self.n_dimensions:
                self.weights[k] *= self.weights_scale
            self.weights_scale = 1.0
        self.timestep += 1
    def get_weights(self):
        cdef int k
        self.hiddenUpdates(None)
        for k from 0<=k<self.n_dimensions:
            self.weights[k] *= self.weights_scale
        self.weights_scale = 1.0
        return self.weights
    def get_dense(self, start, end):
        self.hiddenUpdates(None, start, end)
        w = numpy.zeros(end-start)
        w[:] = self.weights[start:end]
        w *= self.weights_scale
        return w
    def get_weights_l1(self):
        cdef double result = 0.0
        cdef int k
        for k from 0<=k<self.n_dimensions:
            result += fabs(self.weights[k])
        return result * self.weights_scale
    def set_weights(self, w):
        self.weights[:] = w

    def save_binary(self, fname_out):
        numbers = self.get_weights()
        with file(fname_out, 'w') as f:
            numpy.save(f, numbers)

cdef class AvgPer:
    '''
    Simple averaged perceptron
    '''
    # TODO check if this works
    cdef readonly int n_dimensions
    cdef public object fc
    cdef double eta
    cdef readonly double[:] weights
    cdef readonly double[:] avgWeights
    cdef double[:] lastUpdated
    cdef long int timestep

    def __init__(self, n_dimensions):
        self.weights = numpy.zeros(n_dimensions, 'd')
        self.avgWeights = numpy.zeros(n_dimensions, 'd')
        self.lastUpdated = numpy.zeros(n_dimensions, 'd')
        self.timestep = 0
    def set_weights(self, weights):
        if len(weights) != len(self.weights):
            raise ValueError()
        self.weights = weights
        self.avgWeights = self.weights.copy()
    cpdef double score(self, SparseVectorD vec, bint use_avg=True):
        if use_avg and self.timestep > 0:
            return vec._dotFull(& self.avgWeights[0])/self.timestep
        else:
            return vec._dotFull( & self.weights[0])

    def update(self, gradient):
        cdef double a
        cdef int i
        cdef SparseVectorD vec
        cdef double gradf
        cdef double actual_update
        cdef double last_timestep, idle_interval
        cdef coordinate_t feature
        for a, vec in gradient:
            for i from 0 <= i < vec.my_len:
                feature = vec.idx_ptr[i]
                gradf = a * vec.vals_ptr[i]
                last_timestep = self.lastUpdated[feature]
                idle_interval = self.timestep - last_timestep
                self.lastUpdated[feature] = self.timestep
                actual_update = self.weights[feature] + gradf
                self.avgWeights[feature] += self.weights[feature] * idle_interval
                self.weights[feature] = actual_update
        self.timestep += 1

    def __reduce__(self):
        return (make_classifier, ('avgper',
                                  numpy.asarray(self.avgWeights)/self.timestep,
                                  self.fc))
    def save_binary(self, fname_out):
        numbers = numpy.asarray(self.avgWeights)/self.timestep
        with file(fname_out, 'w') as f:
            numpy.save(f, numbers)

cdef class FileModel:
    cdef readonly int n_dimensions
    cdef public object fc
    cdef readonly double[:] weights
    def __init__(self, fname):
        numbers = numpy.load(fname)
        self.weights = numbers
        self.n_dimensions = len(numbers)
    cpdef double score(self, SparseVectorD vec, bint use_avg=True):
        return vec._dotFull(& self.weights[0])
    def get_dense(self, start, end):
        w = numpy.zeros(end-start)
        w[:] = self.weights[start:end]
        return w

def make_classifier(tp, weights, fc=None):
    if tp == 'adagrad':
        result = AdaGrad(len(weights))
        result.set_weights(weights)
        result.fc = fc
        return result
    if tp == 'avgper':
        result = AvgPer(len(weights))
        result.set_weights(weights)
        result.fc = fc
        return result



import sys
import os
from cStringIO import StringIO
from collections import defaultdict
from time import clock, ctime
import traceback
import pdb

rate_limits = {}
next_output = defaultdict(float)
skipped = defaultdict(int)
messages = {}

logfile = sys.stderr

def debug_on_error(tp, value, tb):
    if (not sys.stderr.isatty() or
        not sys.stdin.isatty()):
        log_exception(traceback.format_exception(tp, value, tb))
        original_hook(tp, value, tb)
    else:
        traceback.print_exception(tp, value, tb)
        print
        pdb.pm()

original_hook = sys.excepthook

def install_exception_handler():
    sys.excepthook = debug_on_error

def set_rate_limit(topic, min_sec):
    rate_limits[topic] = min_sec


def set_message(topic, event, s):
    messages[(topic, event)] = s


def add_log_item(topic, event, **kwargs):
    if topic in rate_limits:
        tm = clock()
        if tm < next_output[topic]:
            skipped[topic] += 1
            return
    else:
        tm = None
    if (topic, event) in messages:
        s = messages[(topic, event)] % kwargs
    else:
        s = '[%s %s]' % (
            event,
            ' '.join(['%s:%s' % (k, v) for (k, v) in kwargs.iteritems()]))
    n_skipped = skipped[topic]
    if n_skipped:
        s_skipped = ' [skipped %d]' % (n_skipped,)
        skipped[topic] = 0
    else:
        s_skipped = ''
    print >>logfile, "%s %s %s%s" % (ctime(), topic, s, s_skipped)
    logfile.flush()
    if tm is not None:
        next_output[topic] = tm + rate_limits[topic]

def log_exception(exception_text, topic='ERROR'):
    if type(exception_text) == list:
        exception_text = ''.join(exception_text)
    print >>logfile, "%s %s [[[exception"%(ctime(), topic)
    print >>logfile, exception_text
    print >>logfile, "]]]"

def set_logfile(fname):
    global logfile
    logfile.flush()
    logfile = file(fname, 'a')

def set_logfile_prefix(prefix):
    fname = '%s_%s.log'%(prefix, os.getpid())
    set_logfile(fname)

class LongLogEntry:
    def __init__(self, topic='INFO', event='description', **kwargs):
        if (topic, event) in messages:
            s = messages[(topic, event)] % kwargs
        else:
            s = '[%s %s]' % (
                event,
                ' '.join(['%s:%s' % (k, v) for (k, v) in kwargs.iteritems()]))
        self.topic = topic
        self.s = s
        self.kwargs = kwargs
        self.stream = StringIO()
    def __enter__(self):
        return self.stream
    def __exit__(self, type, value, traceback):
        topic = self.topic
        if topic in rate_limits:
            tm = clock()
            if tm < next_output[topic]:
                skipped[topic] += 1
                return
        else:
            tm = None
        n_skipped = skipped[topic]
        if n_skipped:
            s_skipped = ' [skipped %d]' % (n_skipped,)
            skipped[topic] = 0
        else:
            s_skipped = ''
        print >>logfile, "%s %s %s%s [[["%(ctime(), topic, self.s, s_skipped)
        print >>logfile, self.stream.getvalue(),
        print >>logfile, "]]]"



#!/usr/bin/env python
'''
standard setup script. This needs at least numpy and Cython
to be installed.
'''
from setuptools import setup, Extension

class lazy_cythonize(list):
    def __init__(self, callback):
        self._list = None
        self.callback = callback
    def c_list(self):
        if self._list is None:
            self._list = self.callback()
        return self._list
    def __iter__(self):
        return iter(self.c_list())
    def __getitem__(self, ii):
        return self.c_list()[ii]
    def __len__(self):
        return len(self.c_list())

# depending on gcc version, use gnu++0x or c++11
CXX_STD_OPT = '-std=gnu++0x'

def extensions():
    from Cython.Build import cythonize
    import numpy
    incdirs = [numpy.get_include(), 'include', 'pyx_src', 'pyx_src/lurrn']
    ext_modules = [Extension('lurrn.alphabet',
                             ['pyx_src/lurrn/alphabet.pyx'],
                             language='c++',
                             extra_compile_args=[CXX_STD_OPT],
                             include_dirs=incdirs),
                   Extension('lurrn.sparsmat',
                             ['pyx_src/lurrn/sparsmat.pyx',
                              'src/inverse_erf.cpp'],
                             language='c++',
                             include_dirs=incdirs),
                   Extension('lurrn.feature',
                             ['pyx_src/lurrn/feature.pyx'],
                             language='c++',
                             extra_compile_args=[CXX_STD_OPT],
                             include_dirs=incdirs),
                   Extension('lurrn.learn',
                             ['pyx_src/lurrn/learn.pyx'],
                             language='c++',
                             include_dirs=incdirs)]
    return cythonize(ext_modules)

setup(name='Lurrn',
      version='0.7.2',
      description='Simple machine learning library',
      author='Yannick Versley',
      author_email='versley@cl.uni-heidelberg.de',
      ext_modules=lazy_cythonize(extensions),
      entry_points={
      },
      keywords=['machine learning', 'numpy'],
      packages=['lurrn'],
      package_dir={'': 'py_src'},
      install_requires=['setuptools>=17', 'cython>=0.19', 'numpy'],
      requires=['numpy', 'simplejson']
      )
